@@ -18,8 +18,8 @@ pub(crate) fn decode_helper<'a>(
         return Ok(s.into());
     }
 
-    let mut buffer: Vec<u8> = Vec::with_capacity(bytes.len() * 3);
-    // Safety: decode_slice expects buffer.len() >= src.len() * 3
+    let mut buffer: Vec<u8> = Vec::with_capacity(bytes.len() * 3 + 1);
+    // Safety: decode_slice expects buffer.len() >= src.len() * 3 + 1
     let mut ptr = buffer.as_mut_ptr();
 
     // If we wouldn't gain anything from the word-at-a-time implementation, fall
@@ -69,8 +69,8 @@ pub(crate) fn decode_helper_non_ascii<'a>(
     bytes: &'a [u8],
     fallback: Option<char>,
 ) -> Result<Cow<'a, str>, DecodeError> {
-    let mut buffer: Vec<u8> = Vec::with_capacity(bytes.len() * 3);
-    // Safety: decode_slice expects buffer.len() >= src.len() * 3
+    let mut buffer: Vec<u8> = Vec::with_capacity(bytes.len() * 3 + 1);
+    // Safety: decode_slice expects buffer.len() >= src.len() * 3 + 1
     let mut ptr = buffer.as_mut_ptr();
     let fallback: Option<UTF8Entry> = fallback.map(UTF8Entry::from_char);
     unsafe { decode_slice(table, &mut ptr, bytes, fallback) }?;
@@ -80,7 +80,7 @@ pub(crate) fn decode_helper_non_ascii<'a>(
 /// Lookup every byte in [`src`] using provided [`table`] and write resulting bytes to [`ptr`]
 /// # Safety
 ///
-/// This function is unsafe because it assumes that the buffer pointed to by [`ptr`] has a length >= src.len() * 3
+/// This function is unsafe because it assumes that the buffer pointed to by [`ptr`] has a length >= src.len() * 3 + 1
 #[inline]
 unsafe fn decode_slice(
     table: &Table,
@@ -89,11 +89,14 @@ unsafe fn decode_slice(
     fallback: Option<UTF8Entry>,
 ) -> Result<(), DecodeError> {
     for (i, b) in src.iter().enumerate() {
-        let UTF8Entry { buf, len } = table[*b as usize].or(fallback).ok_or(DecodeError {
-            position: i,
-            value: *b,
-        })?;
-        ptr.copy_from_nonoverlapping(buf.as_ptr(), 3);
+        let entry @ UTF8Entry { buf: _, len } =
+            table[*b as usize].or(fallback).ok_or(DecodeError {
+                position: i,
+                value: *b,
+            })?;
+        // Safety: size of entry is 4 bytes starting with 3 bytes of UTF8
+        // ptr is only moved 1-3 bytes
+        (*ptr as *mut u32).write(std::mem::transmute(entry));
         *ptr = ptr.add(len as usize);
     }
     Ok(())
